# encoding: utf-8
# cython: profile=True
# icc_2d_2.py
#
# Supplies functions to perform two-dimensional inelastic close-coupling 
# calculations for scattering He atoms from corrugated surfaces.
#
# Patrick Kraus
# 06.02.14
# 
# Michael Pusterhofer
# 31.10.2016

################################################################################
# Import libraries
################################################################################
from scipy.special import iv
from haslib.science_basics import *
import numpy as np
import itertools as it
import sys

################################################################################
# Import cython libraries
################################################################################
cimport fox_goodwin as fg
cimport numpy as np
cimport cython

def iCC_2D_2(h, AP, a, D, xi, Ei, theta_i, G1, G2,
                Gope = -1, Gclo=0, phon = [0,0,0], floq = 1, sta_z = -7.0*A,
                end_z = 20.0*A, st_p_wl = 150, CC_border = 1E-7,
                write_P = False, verbose = False, m = m_He3, maxrad=20):
    """
    Two-dimensional calculation of scattering intensities, taking inelastic contributions
    into account.
    The CC-calculation is done using a Fox-Goodwin integrator
    The potential function is assumed to be a Corrugated Morse shape
    
    Patrick Kraus
    06.Feb.2014
    email: patrick.kraus@tugraz.at
    
    Args:
        h: Corrugation Parameter
        AP: Inelastic coupling Parameter
        a: 1D lattice constant (in meter)
        D: Potential depth (in joule)
        xi: Potential stiffnes of the central morse potential (in m**-1)
        Ei: Incident helium energy (in joule)
        theta_i: Incident angle of the scattering (in degrees)
        G1: Reciprocal lattice vector 1 (in m**-1)
        G2: Reciprocal lattice vector 2 (in m**-1)
        Gope: Number of open channels to be considered - (Standard Value "-1" - all the channels)
        Gclo: Number of closed channels to be considered - (Standard Value "0")
        phon: Array carrying the phonon values array([Kval_x, Kval_y, Eval])
        floq: Number of Floquet Blocks on both sides - default value is 1
        sta_z: starting value of the numerov interation (in meter) - (Standard Value "7.0*A")
        end_z: terminal value of the numerov interation (in meter) - (Standard Value "20.0*A")
        st_p_wl: Number of steps to be performed in the interation within the length of the shortest calculated wavelength - (Standard Value "150")
        CC_border: Minimum value for the coupling parameters to be considered - (Standard Value "1E-7")
        write_P: Boolean Flag to determine if the wavefunctions will be written and returned (Standard Value - "False")
        verbose: Boolean Flag to signal if Parameter Information should be displayed - (Standard Value "False")
    
    Results:
        array: Scattering Intensities sorted by channels and Floquet Blocks (basically by energy)

        ::

            [[Gx1  , Gx2  , Gx3  , ...],
             [Gy1  , Gy2  , Gy3  , ...],
             [F1   , F2   , F3   , ...],
             [Kx1  , Kx2  , Kx3  , ...],
             [Ky1  , Ky2  , Ky3  , ...],
             [Kz2_1, Kz2_2, Kz2_3, ...],
             [I1   , I2   , I3   , ...]]

        (Only considering open channels)

        or 
        Array, holding the wavefunctions of the open channels in its rows, if
        the "write_P" Flag is set.
    
    References:
        .. [1] A.S. Sanz and S. Miret-Artes - Selective Adsorption Resonances: Quantum And Stochastic Approaches, Phys. Rep. 451 (2007), 37-154
        .. [2] S. Miret-Artes - Multiphonon contributions to diffracted intensities and resonance profiles in atom-surface scattering, Surf. Sci. 366 (1996), L681-L684
        .. [3] S. Miret-Artes - Resonant inelastic scattering of atoms from surfaces, Surf. Sci. 339 (1995), 205-220
    """
    ############################################################################
    # Channel calculation
    ############################################################################
    
    # Phonon properties
    phon_K = np.array([phon[0],phon[1]])
    phon_E = phon[2]

    # Kinematical properties
    ki = np.sqrt(Ei/h2m(m))
    Ki = np.array([ki*np.sin(theta_i*np.pi/180.0)])

    # Determine Number of Floquet Blocks considered (including the 0 Block)
    Nr_fl = 2 * floq + 1

    # Calculation of maximal distance for G-vectors to be considered
    maxrad2 = 4 + int(round(h, 1) * 10)    
    maxrad = max(maxrad, maxrad2)

    # Now we produce a list of G-vectors to be checked for relevance
    grid = np.zeros([Nr_fl*(2*maxrad+3)**2,5])
    range_g = np.arange(2*maxrad+3) - (maxrad+1)
    range_floq = np.arange(-floq, floq+1, 1)
    lst = np.array(list(it.product(range_g, range_g, range_floq)))
    grid[:,[0,1]] = lst[:,0,None]*G1[[0,1]] + \
                    lst[:,1,None]*G2[[0,1]] + \
                    lst[:,2,None]*phon_K
    grid[:,[2,3,4]] = lst[:,[0,1,2]]

    # Now determine the distances of the channels
    # ignore everything farther away than maxrad
    dist = np.sqrt(grid[:,0]**2 + grid[:,1]**2)
    flag = dist <= maxrad * max(np.linalg.norm(G1),np.linalg.norm(G2))    

    # Produce a container with the chosen values
    container = np.zeros([sum(flag), 7])
    container[:,0:5] = grid[flag,:]
    Ki_x_new = Ki + container[:,0]
    Ki_y_new = container[:,1]

    # Calculate final energy
    container[:,5] = Ki_x_new
    container[:,6] = ki**2 - Ki_x_new**2 - Ki_y_new**2 - \
                     container[:,4]*phon_E/h2m(m)
    
    # Now sort the channels kz2-wise (highest = [0])
    I = np.argsort(container[:,6])[::-1]
    container = container[I,:]    
    
    # seperate into open and closed channels
    chan_open = container[container[:,6]>=0,:]
    chan_clos = container[container[:,6]<0,:]
    
    # Now choose only the channels within the range given by the starting parameters
    if Gope==-1 or Gope > chan_open.shape[0]:
        pass
    else:
        #select Gope lowest open channels
        choice = -Gope
        chan_open = chan_open[choice:,:]
        
    if Gclo==-1 or Gclo>chan_clos.shape[0]:
        pass
    else:
        #select Gclo highest closed channels
        chan_clos = chan_clos[:Gclo,:]
    
    # Read some simulation parameters resulting from those limits
    Nr_open = chan_open.shape[0]
    Nr_clos = chan_clos.shape[0]
    Nr_C = Nr_open+Nr_clos
    
    # Find index of specular channel
    spe_c = np.where((chan_open[:,2] == 0) & 
            (chan_open[:,3] == 0) &
            (chan_open[:,4] == 0))[0]
        
    # Combine closed and open containers again
    channels = np.vstack([chan_open,chan_clos])
    
    # Calculate the numerov step
    step = 2 * np.pi / (st_p_wl * np.sqrt(channels[:,6].max()))
    
    # Definition of the z-Step list
    z = np.arange(sta_z, end_z, step)
    NrS = z.size
    
    # Time for the first information piece for the user
    if verbose:
        print("""
#-------------------------------------------------
iCC_2D v0.5 - 11.Feb.2014 - Patrick Kraus
including elastic Channel integration
#-------------------------------------------------

Channel Information:
    Total Number of Channels               -->   {0}
    Number of Floquet Blocks               -->   {1}
    Number of reciprocal Lattice Positions -->   {2}
    
    Open channels                          -->   {3}
    Closed channels                        -->   {4}
    
Integration Information:
    Number of Steps                        -->   {5}
    Number of Steps per minimum wavelength -->   {6}
    Length of Step (A)                     -->   {7}
""".format(Nr_C,Nr_fl,np.floor(Nr_C/Nr_fl).astype(int),Nr_open,Nr_clos,NrS,st_p_wl,step/A))
    
    ############################################################################
    # Coupling calculation
    ############################################################################

    # Now a function to calculate the matrix for the system (at z==0)
    # This matrix can stay constant throughout the calculation. To determine
    # the z-dependence we only need to apply a multiplicative factor (for CMP)
    Mas_0 = np.zeros([Nr_C,Nr_C])    # <-- will hold markers for specular contributions
    Mas_F0 =np.zeros([Nr_C,Nr_C])    # <-- will hold markers for inelastic specular contributions
    Mas_K = np.zeros([Nr_C,Nr_C])    # <-- will hold markers for K2z values on specular contributions
    Mas_E = np.zeros([Nr_C,Nr_C])    # <-- will hold the asymptotic energy of spec. contrib.

    max_dg1 = int(max(abs(channels[:,2])))
    max_dg2 = int(max(abs(channels[:,3])))
    max_dg1 =max_dg2 = np.array([max_dg1,max_dg2]).max()
    max_df = int(max(abs(channels[:,4])))
    Coupl = -1*np.ones([4*max_dg1+1,4*max_dg2+1,4*max_df+1])
    V_0 = two_D_el_coupling(a,h,xi,0,0)
    
    np.fill_diagonal(Mas_K,channels[:,6])
    Mas_0 = np.eye(Nr_C,Nr_C)
    for ind_l in range(Nr_C):
        for ind_c in range(Nr_C):
            if (channels[ind_l,2]==channels[ind_c,2])&(channels[ind_l,3]==channels[ind_c,3]):
                pass
                if (channels[ind_l,4]==channels[ind_c,4]):
                    # Elastic specular contrib - mark Mas_0 and Mas_K
                    Mas_0[ind_l,ind_c] = 1
                    Mas_K[ind_l,ind_c] = channels[ind_l,6]
                elif abs(channels[ind_l,4]-channels[ind_c,4])==1:
                    pass
                    # Inelastic specular contribution - mark Mas_F0
                    Mas_F0[ind_l,ind_c] = 1
            else:
                #reduce couplings evaluation!!
                dG1 = int(channels[ind_l,2]-channels[ind_c,2])
                dG2 = int(channels[ind_l,3]-channels[ind_c,3])
                dF = int(channels[ind_l,4]-channels[ind_c,4])
                coupl = Coupl[  dG1+2*max_dg1,
                                dG2+2*max_dg2,
                                dF+2*max_df]
                if coupl < 0:
                    coupl = couplings(dG1,dG2,dF,h,a,xi,AP,CC_border,V_0)  
                    Coupl[ dG1+2*max_dg1, dG2+2*max_dg2,dF+2*max_df] = coupl
                    Coupl[-dG1+2*max_dg1,-dG2+2*max_dg2,dF+2*max_df] = coupl
                    Coupl[-dG2+2*max_dg1,-dG1+2*max_dg2,dF+2*max_df] = coupl
                    Coupl[ dG2+2*max_dg1, dG1+2*max_dg2,dF+2*max_df] = coupl

                Mas_E[ind_l,ind_c] = coupl
    
    
    # define a function to supply the coupling matrix for every z
    def  giveme_w(np.float64_t z,np.uint16_t z_index):
        W = np.zeros_like(Mas_0)
        # Define specular contributions
        W = Mas_0*centr_pot(z,D,xi)/h2m(m) - Mas_K
        # Define inelastic specular contributions (ignored a.t.m.)
        #ISP = 0*Mas_F0#*centr_pot_dev(z,D,xi)/h2m(m)
        # Weigh couplings with the distance
        W += Mas_E*D*np.exp(-2*xi*z)/h2m(m)
        
        # return complete coupling matrix
        return W
    
    K2z = channels[:,6]
    
    ############################################################################
    # Numerov Routine 
    ############################################################################

    # Preparation of used containers
    R_00 = np.zeros([Nr_C,Nr_C]) # Filling it up with zeros - seemed to work
    R_00 = fg.fox_goodwin(z, giveme_w,write_P)

    ############################################################################
    # K-Extraction
    ############################################################################

    # Calculation of the S0 (Sinus-Zero) and the CE (Cosinus-Exponential) - Vectors
    # for the last two z-positions

    S0_M1 = np.zeros([Nr_C,1])
    S0_00 = np.zeros_like(S0_M1)
    CE_M1 = np.zeros_like(S0_M1)
    CE_00 = np.zeros_like(S0_M1)
    
    run = NrS-2 

    # If the normal energy is positive, the corresponding sin and cos
    # values are calculated and set
    ind1 = K2z>0
    kz = np.sqrt(K2z[ind1,None])
    S0_M1[ind1] = np.sin(kz*z[run])/np.sqrt(kz)
    S0_00[ind1] = np.sin(kz*z[run+1])/np.sqrt(kz)
    CE_M1[ind1] = np.cos(kz*z[run])/np.sqrt(kz)
    CE_00[ind1] = np.cos(kz*z[run+1])/np.sqrt(kz)

    # If the normal energy is negative, only the CE values are set for
    # the use inside the decaying (hopefully) exponentials
    ind2 = K2z<=0
    kz = np.sqrt(-K2z[ind2,None])
    CE_M1[ind2] = np.exp(-kz*z[run])/np.sqrt(kz)
    CE_00[ind2] = np.exp(-kz*z[run+1])/np.sqrt(kz)
          
    S0_M1 = np.diagflat(S0_M1)
    S0_00 = np.diagflat(S0_00)
    CE_M1 = np.diagflat(CE_M1)
    CE_00 = np.diagflat(CE_00)
    
    # Now everything should be prepared and we can solve the simple matrix
    # equation
    SS = S0_M1 - np.dot(R_00,S0_00)
    CC = np.dot(R_00,CE_00) - CE_M1

    KK = np.linalg.solve(CC,SS)
    # Here the values of K and E are still mixed
    # The values belonging to closed channels in KK are removed
    geschl = np.where(K2z<0)[0]
    K = np.delete(np.delete(KK,geschl,0),geschl,1)
    
    NUM = K.shape[0]
    S = np.linalg.solve((np.eye(NUM)-(0+1j)*K),np.eye(NUM)+(0+1j)*K)
    S = abs(S)**2
    
    opench = np.where(K2z>0)[0]
    CHA_x = channels[opench,2].reshape([-1,1])
    CHA_y = channels[opench,3].reshape([-1,1])
    FLO = channels[opench,4].reshape([-1,1])
    Kval_x = channels[opench,5].reshape([-1,1])
    Kval_y = channels[opench,1].reshape([-1,1])
    K2z = K2z[opench].reshape([-1,1])
    
    # Return the prepared values
    return np.hstack([CHA_x,CHA_y,FLO,Kval_x,Kval_y,K2z,S[spe_c,:].reshape([-1,1])])
    
# -----------------------------------------------------------------------------         
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

cdef np.float64_t two_D_el_coupling(np.float64_t a,np.float64_t h,np.float64_t xi,np.int16_t G1,np.int16_t G2,np.float64_t div=1.0):
    """
    calc_coupl_2d_hex(a,h,G1,G2,div=1.0):
    
    Input variables:
        a   ... Edge length of the primitive cell hexagon (in meters)
        h   ... Corrugation Parameter of the cosine corrugation function
        xi  ... Stiffnes parameter of the corrugated morse parameter (in m**-1)
        G1  ... First index of G-Vector to the channel
        G2  ... Second index of G-Vector to the channel
        div ... divison parameter - The solution will be divided by this number
    
    Output variables:
        V_G ... Coupling factor from (0,0) to (G1,G2)
    """
    cdef int kmax = 10 
    cdef np.float64_t alpha
    cdef np.ndarray[np.int8_t,ndim=1] k = np.zeros(2 * kmax + 1,dtype=np.int8)
    cdef np.ndarray[np.float64_t,ndim=1] qu = np.zeros(2 * kmax + 1)
    
    k = np.arange(-1*kmax,kmax+1,1,dtype=np.int8)
    alpha = 2*xi*h*a/3.0
    qu = iv(k,alpha)*( iv(k+G2,alpha)*iv(k-G1,alpha) + iv(k-G2,alpha)*iv(k+G1,alpha) )/div
    
    return np.sum(qu)

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

# Define a function for the inter- and intrablock couplings
cdef np.float64_t couplings(np.int16_t dGx,np.int16_t dGy,np.int16_t dF, np.float64_t h, np.float64_t a, np.float64_t xi, np.float64_t B, np.float64_t CC_border,np.float64_t V_0):
    cdef np.float64_t V_G
    # Now check for multi-phonon interaction
    if abs(dF)>1:
        return 0
    # for all the other cases we can calculate          
    V_G = two_D_el_coupling(a,h,xi,dGx,dGy,V_0)
    # Now we only need to check if the interaction is elastic or inelastic
    if dF==0:
        # Elastic case
        if V_G > CC_border:
            return V_G
        else:
            return 0
    else:
        # inelastic case
        if V_G > CC_border:
            return -2*B*xi*V_G
        else:
            return 0

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

cdef np.float64_t centr_pot(np.float64_t z,np.float64_t D,np.float64_t xi):
    # Define central interaction potential (Morse?)
    return D*( np.exp(-2*xi*z) - 2*np.exp(-xi*z) )

def  centr_pot_dev(z,D,xi):
    return 2*D*xi*(np.exp(-xi*z) - np.exp(-2*xi*z))

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

if __name__=="__main__":
    a = 4.3084*A
    Sb111 = Crystal(a,a,120)         # Define Sb structure
    Sb111.rotate_crystal(1,0)        # Define scattering direction
    Gx = Sb111.b1[0:2]
    Gy = Sb111.b2[0:2]
    for i in range(10):
        ERG = iCC_2D(0.1,0,4.308*A,4.196*mel,0.38/A,13*mel,45.75,Gx,Gy,Gope=-1,Gclo=2,phon=np.array([0,0,0]),floq=0,sta_z=-7.0*A,end_z=20.0*A,st_p_wl=100,CC_border=1E-4,verbose=True)
    print(ERG)
